from enum import Enum

class TypeCategory(Enum):
    ENUM = 2
    FLAG = 3
    STRUCTURE = 4
    TUPLE = 20
    BUILTIN = 99

class TypeSpecifier:
    name: str
    category: TypeCategory
    library: str | None
    register: bool
    def __init__(self, name: str, category: TypeCategory) -> None: ...



from typing import Callable

from mdk.types.expressions import Expression
from mdk.types.specifier import TypeSpecifier

__all__ = ['Abs', 'Acos', 'AiLevel', 'Alive', 'Anim', 'AnimElemNo', 'AnimElemTime', 'AnimExist', 'AnimTime', 'Asin', 'Atan', 'AuthorName', 'BackEdgeBodyDist', 'BackEdgeDist', 'CanRecover', 'Ceil', 'Command', 'Const', 'Const240p', 'Const480p', 'Const720p', 'Cos', 'Ctrl', 'DrawGame', 'E', 'Exp', 'Facing', 'Floor', 'FrontEdgeBodyDist', 'FrontEdgeDist', 'GameHeight', 'GameTime', 'GameWidth', 'GetHitVar', 'HitCount', 'HitFall', 'HitOver', 'HitPauseTime', 'HitShakeOver', 'ID', 'InGuardDist', 'IsHelper', 'IsHomeTeam', 'Life', 'LifeMax', 'Ln', 'Log', 'Lose', 'MatchNo', 'MatchOver', 'MoveContact', 'MoveGuarded', 'MoveHit', 'MoveReversed', 'Name', 'NumEnemy', 'NumExplod', 'NumHelper', 'NumPartner', 'NumProj', 'NumProjID', 'NumTarget', 'P1Name', 'P2Life', 'P2StateNo', 'P2Name', 'P3Name', 'P4Name', 'PalNo', 'Pi', 'Power', 'PowerMax', 'PlayerIDExist', 'ProjCancelTime', 'ProjContactTime', 'ProjGuardedTime', 'ProjHitTime', 'Random', 'RoundNo', 'RoundsExisted', 'RoundState', 'SelfAnimExist', 'Sin', 'StateNo', 'StageVar', 'Tan', 'TeamSide', 'TicksPerSecond', 'Time', 'Win', 'WinKO', 'WinTime', 'WinPerfect', 'HitDefAttr', "StateType", "MoveType", "TeamMode", "P2MoveType", "P2StateType", "HitVel", "P2BodyDist", "P2Dist", "ParentDist", "Pos", "RootDist", "ScreenPos", "Vel", "ifelse", "Cond", "EmptyExpression", "EmptyTuple", "AnimElem", "ScreenWidth", "ScreenHeight"]

class PositionExpression:
    x: Expression
    y: Expression
    def __init__(self, name: str, type: TypeSpecifier): ...

Abs: Callable[..., Expression]
Acos: Callable[..., Expression]
AiLevel: Expression
Alive: Expression
Anim: Expression
AnimElemNo: Callable[..., Expression]
AnimElemTime: Callable[..., Expression]
AnimExist: Callable[..., Expression]
AnimTime: Expression
Asin: Callable[..., Expression]
Atan: Callable[..., Expression]
AuthorName: Expression
BackEdgeBodyDist: Expression
BackEdgeDist: Expression
CanRecover: Expression
Ceil: Callable[..., Expression]
Command: Expression
Cond: Callable[..., Expression]
Const: Callable[..., Expression]
Const240p: Callable[..., Expression]
Const480p: Callable[..., Expression]
Const720p: Callable[..., Expression]
Cos: Callable[..., Expression]
Ctrl: Expression
DrawGame: Expression
E: Expression
Exp: Callable[..., Expression]
Facing: Expression
Floor: Callable[..., Expression]
FrontEdgeBodyDist: Expression
FrontEdgeDist: Expression
GameHeight: Expression
GameTime: Expression
GameWidth: Expression
GetHitVar: Callable[..., Expression]
HitCount: Expression
HitDefAttr: Expression
HitFall: Expression
HitOver: Expression
HitPauseTime: Expression
HitShakeOver: Expression
HitVel: PositionExpression
ID: Expression
ifelse: Callable[..., Expression]
InGuardDist: Expression
IsHelper: Callable[..., Expression]
IsHomeTeam: Expression
Life: Expression
LifeMax: Expression
Ln: Callable[..., Expression]
Log: Callable[..., Expression]
Lose: Expression
MatchNo: Expression
MatchOver: Expression
MoveContact: Expression
MoveGuarded: Expression
MoveHit: Expression
MoveReversed: Expression
Name: Expression
NumEnemy: Expression
NumExplod: Callable[..., Expression]
NumHelper: Callable[..., Expression]
NumPartner: Expression
NumProj: Expression
NumProjID: Callable[..., Expression]
NumTarget: Callable[..., Expression]
P1Name: Expression
P2BodyDist: PositionExpression
P2Dist: PositionExpression
P2Life: Expression
P2StateNo: Expression
P2Name: Expression
P3Name: Expression
P4Name: Expression
PalNo: Expression
ParentDist: PositionExpression
Pi: Expression
Pos: PositionExpression
Power: Expression
PowerMax: Expression
PlayerIDExist: Callable[..., Expression]
PrevStateNo: Expression
ProjCancelTime: Callable[..., Expression]
ProjContactTime: Callable[..., Expression]
ProjGuardedTime: Callable[..., Expression]
ProjHitTime: Callable[..., Expression]
Random: Expression
RoundNo: Expression
RoundsExisted: Expression
RoundState: Expression
RootDist: PositionExpression
ScreenPos: PositionExpression
SelfAnimExist: Callable[..., Expression]
Sin: Callable[..., Expression]
StateNo: Expression
StageVar: Callable[..., Expression]
Tan: Callable[..., Expression]
TeamSide: Expression
TicksPerSecond: Expression
Time: Expression
Vel: PositionExpression
Win: Expression
WinKO: Expression
WinTime: Expression
WinPerfect: Expression
StateType: Expression
MoveType: Expression
TeamMode: Expression
P2MoveType: Expression
P2StateType: Expression
HitVel: PositionExpression
P2BodyDist: PositionExpression
P2Dist: PositionExpression
ParentDist: PositionExpression
Pos: PositionExpression
RootDist: PositionExpression
ScreenPos: PositionExpression
Vel: PositionExpression
AnimElem: Expression
ScreenWidth: Expression
ScreenHeight: Expression

EmptyExpression: Expression
EmptyTuple: tuple[Expression]


## this is a utility for viewing animations built via MDK (code-based animations).
import os
import sys
import math
import importlib.util
from pathlib import Path

from mdk.types.context import CompilerContext
from mdk.resources.animation import Animation, Sequence, SequenceModifier, Frame

from mdk.utils.sff import SFF

from PyQt6.QtCore import Qt, QFileSystemWatcher
from PyQt6.QtGui import QImage, QPixmap, QTransform
from PyQt6.QtWidgets import (
    QApplication,
    QMainWindow,
    QGridLayout,
    QWidget,
    QComboBox,
    QLabel,
    QTabWidget,
    QTextEdit,
    QPushButton,
    QSlider
)

class MainWindow(QMainWindow):
    def __init__(self, input_file: str, sff: SFF):
        super().__init__()
        self._loaded = False
        self.sff = sff

        directory = os.path.dirname(os.path.abspath(input_file))
        ## this is hacky, but i guess most venv will follow this name?
        watches = [str(file) for file in Path(directory).rglob("*.py") if file.is_file() and "venv" not in str(file.parent)]
        print(f"Watching files for modifications: {watches}")
        self.watcher = QFileSystemWatcher(watches)
        self.watcher.fileChanged.connect(self.onModuleUpdated)

        self.path = input_file
        self.packages = [k for k in sys.modules]
        self.module = isolate(self.path)
        self.components = recurse_animations(self.module, [])

        self.setWindowTitle("MDK Animation Viewer")
        self.resize(1200, 800)

        center = QWidget()
        self.setCentralWidget(center)

        self.gridLayout = QGridLayout()
        center.setLayout(self.gridLayout)

        ## upper left: display image
        image = QImage(400, 400, QImage.Format.Format_RGBA8888)
        image.fill(Qt.GlobalColor.black)

        pixmap = QPixmap.fromImage(image)

        self.imageLabel = QLabel()
        self.imageLabel.setPixmap(pixmap)

        self.gridLayout.addWidget(self.imageLabel, 0, 0, 5, 1, Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)

        ## upper right: select component
        self.componentSwitch = QComboBox()
        self.componentSwitch.addItems([c['name'] for c in self.components])
        self.componentSwitch.currentIndexChanged.connect(self.onAnimSelected)
        self.gridLayout.addWidget(self.componentSwitch, 0, 1)

        ## middle right: toggles and reload
        reload = QPushButton()
        reload.clicked.connect(self.onModuleUpdated)
        reload.setText("Reload")
        self.gridLayout.addWidget(reload, 1, 1)

        ## middle right: frame select
        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.slider.valueChanged.connect(self.onFrameSelected)
        self.gridLayout.addWidget(self.slider, 2, 1)

        ## middle right: frame number, total frames, animation length
        self.animtext = QLabel()
        self.animtext.setText("Loading...")
        self.gridLayout.addWidget(self.animtext, 3, 1)

        ## lower right: infobox
        self.infobox = QTabWidget()
        self.pythonContent = QTextEdit()
        self.pythonContent.setReadOnly(True)
        self.cnsContent = QTextEdit()
        self.cnsContent.setReadOnly(True)
        self.infobox.addTab(self.pythonContent, "Python (Decompiled)")
        self.infobox.addTab(self.cnsContent, "AIR (Output)")
        self.gridLayout.addWidget(self.infobox, 4, 1)

        ## trigger onAnimSelected once to preload python/CNS/sprites
        self.onAnimSelected()

        self._loaded = True

    def onAnimSelected(self):
        ## disconnect the slider events first!
        self.slider.valueChanged.disconnect(self.onFrameSelected)

        ## note the weird try/excepts here are to preserve id of None
        ## (since `Animation.compile()` assigns IDs to animations in the context)
        component = self.components[self.componentSwitch.currentIndex()]['name']
        script: Animation = self.components[self.componentSwitch.currentIndex()]['obj']

        content = script.python(component)
        self.pythonContent.setText(content)

        try:
            original_id = script._id
        except:
            pass

        content = script.compile()
        if content.startswith("[Begin Action"):
            content = "\n".join(content.split("\n")[1:])
        self.cnsContent.setText(content)

        try:
            script._id = original_id # type: ignore
        except:
            pass

        self.frames: list[Frame] = script.sequence._frames if isinstance(script, Animation) else script._frames

        self.slider.setMinimum(0)
        self.slider.setMaximum(len(self.frames) - 1)
        self.slider.setValue(0)

        self.onFrameSelected()

        self.slider.valueChanged.connect(self.onFrameSelected)

    def onFrameSelected(self):
        frame = self.slider.value()
        sum_len = sum([x._length for x in self.frames])
        sum_now = sum([x._length for x in self.frames[:frame]])
        self.reload_sprite(frame, 0)
        self.animtext.setText(f"Frame {frame + 1} / {len(self.frames)}: tick {sum_now} / {sum_len}, length {self.frames[frame]._length}")
    
    def reload_sprite(self, index: int, palette: int):
        sprite = self.sff.find_sprite(self.frames[index]._group, self.frames[index]._index)
        if sprite != None:
            temp_paletted = sprite.palette(self.sff.palettes[palette])
            scale_ratio =  self.frames[index]._scale[0] / self.frames[index]._scale[1]
            if scale_ratio > 0:
                scale_width = 400
                scale_height = math.ceil(400 / scale_ratio)
            else:
                scale_width = math.ceil(400 / scale_ratio)
                scale_height = 400
            image = QImage(temp_paletted, sprite.width, sprite.height, QImage.Format.Format_RGBA8888).scaled(scale_width, scale_height)
            if self.frames[index]._rotate != 0:
                transform = QTransform()
                transform.rotate(self.frames[index]._rotate * -1)
                image = image.transformed(transform).scaled(scale_width, scale_height)
            pixmap = QPixmap.fromImage(image)
            self.imageLabel.setPixmap(pixmap)
        else:
            image = QImage(400, 400, QImage.Format.Format_RGBA8888)
            image.fill(Qt.GlobalColor.black)
            pixmap = QPixmap.fromImage(image)
            self.imageLabel.setPixmap(pixmap)

    def onModuleUpdated(self):
        selection = self.components[self.componentSwitch.currentIndex()]

        ## discard the loaded module and reload from scratch.
        try:
            ## remove the old context, since it will contain a full list of animations from prior loads.
            del CompilerContext._instance
            self.module = isolate(self.path)
            reloads: list[str] = []
            for k in sys.modules:
                if k not in self.packages:
                    print(f"Hot reload dynamic module {k}")
                    reloads.append(k)
            for k in reloads:
                importlib.reload(sys.modules[k])
            self.components = recurse_animations(self.module, [])

            self.componentSwitch.currentIndexChanged.disconnect()
            self.componentSwitch.clear()
            self.componentSwitch.addItems([c['name'] for c in self.components])
            self.componentSwitch.currentIndexChanged.connect(self.onAnimSelected)
            idx = next((index for index, d in enumerate(self.components) if d["name"] == selection['name']), None)
            if idx != None:
                self.componentSwitch.setCurrentIndex(idx)
            else:
                self.componentSwitch.setCurrentIndex(0)
            self.reload_sprite(0, 0)
        except Exception as exc:
            self.cnsContent.setText(f"Failed to load module!\n{exc}")
            self.pythonContent.setText(f"Failed to load module!\n{exc}")

def isolate(path):
    if not os.path.dirname(os.path.abspath(path)) in sys.path:
        sys.path.append(os.path.dirname(os.path.abspath(path)))
    # isolated namespace
    spec = importlib.util.spec_from_file_location("isolate", path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not create module spec for {path}")
    
    # import module in namespace
    module = importlib.util.module_from_spec(spec)
    
    # execute
    spec.loader.exec_module(module)
    
    return module

def recurse_animations(module, visited: list[str] = []):
    components = []
    for key in module.__dict__:
        if isinstance(module.__dict__[key], Animation) or isinstance(module.__dict__[key], Sequence) or isinstance(module.__dict__[key], SequenceModifier):
            if not any([x for x in components if x['name'] == key]):
                components.append({ "name": key, "obj": module.__dict__[key] })
        elif type(module.__dict__[key]).__name__ == "module" and key not in visited:
            visited.append(key)
            for c in recurse_animations(module.__dict__[key], visited):
                if not any([x for x in components if x['name'] == c['name']]):
                    components.append(c)

    return components

def launch():
    if len(sys.argv) < 3:
        raise Exception("Usage: mdkair <build script> <sff file>")
    input_file = sys.argv[1]
    sff_file = sys.argv[2]

    ## the window falls in this context manager because `sff_file` needs to stay open.
    with open(sff_file, mode='rb') as f:
        sff = SFF.load(f)

        app = QApplication(sys.argv)
        window = MainWindow(input_file, sff)
        window.show()
        app.exec()

if __name__ == "__main__":
    launch()


from mdk.types.context import StateController, StateScope, TranslationMode
from mdk.types.expressions import Expression
from mdk.types.specifier import TypeSpecifier
from mdk.types.enums import StateType, MoveType, PhysicsType
from mdk.resources.animation import Animation
from typing import Callable

__all__ = ['build', 'library', 'statedef', 'create_statedef', 'template', 'trigger']

def build(def_file: str, output: str, run_mtl: bool = True, skip_templates: bool = False, locations: bool = True, compress: bool = False, preserve_ir: bool = False, target_folder: str = "mdk-out", debug_build: bool = False) -> None: ...
def library(inputs: list[Callable[..., None]], dirname: str = '', output: str | None = None, locations: bool = True) -> None: ...
def statefunc(mode: TranslationMode = TranslationMode.STANDARD) -> Callable[[Callable[..., None]], Callable[..., None]]: ...
def statedef(type: StateType | None = ..., movetype: MoveType | None = ..., physics: PhysicsType | None = ..., anim: int | Animation | None = None, velset: tuple[float, float] | None = None, ctrl: bool | None = None, poweradd: int | None = None, juggle: int | None = None, facep2: bool | None = None, hitdefpersist: bool | None = None, movehitpersist: bool | None = None, hitcountpersist: bool | None = None, sprpriority: int | None = None, stateno: int | None = None, scope: StateScope | None = None, mode: TranslationMode = TranslationMode.STANDARD) -> Callable[[Callable[[], None]], Callable[..., StateController]]: ...
def create_statedef(fn: Callable[[], None], type: Expression = ..., movetype: Expression = ..., physics: Expression = ..., anim: int | Animation | None = None, velset: tuple[float, float] | None = None, ctrl: bool | None = None, poweradd: int | None = None, juggle: int | None = None, facep2: bool | None = None, hitdefpersist: bool | None = None, movehitpersist: bool | None = None, hitcountpersist: bool | None = None, sprpriority: int | None = None, stateno: int | None = None, scope: StateScope | None = None, mode: TranslationMode = TranslationMode.STANDARD) -> Callable[..., StateController]: ...
def template(inputs: list[TypeSpecifier] = [], library: str | None = None, validator: Callable[..., dict[str, Expression]] | None = None) -> Callable[[Callable[..., None]], Callable[..., StateController]]: ...
def trigger(inputs: list[TypeSpecifier], result: TypeSpecifier, library: str | None = None, validator: Callable[..., dict[str, Expression]] | None = None) -> Callable[[Callable[..., Expression]], Callable[..., Expression]]: ...

class ControllerProps:
    def __init__(self, ignorehitpause: Expression | bool | None = None, persistent: Expression | int | None = None): ...
    def __enter__(self) -> ControllerProps: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> bool: ...


from mdk.types.specifier import TypeSpecifier

__all__ = ['AnyType', 'IntType', 'FloatType', 'BoolType', 'ShortType', 'ByteType', 'CharType', 'StringType', 'StateNoType', 'AnimType', 'SoundType', 'SpriteType']

AnyType: TypeSpecifier
IntType: TypeSpecifier
FloatType: TypeSpecifier
BoolType: TypeSpecifier
ShortType: TypeSpecifier
ByteType: TypeSpecifier
CharType: TypeSpecifier
StringType: TypeSpecifier
UStringType: TypeSpecifier
StateNoType: TypeSpecifier
AnimType: TypeSpecifier
SoundType: TypeSpecifier
SpriteType: TypeSpecifier



from mdk.types.expressions import Expression, ConvertibleExpression
from mdk.stdlib.triggers import PositionExpression
from typing import Callable, Optional, Protocol

__all__ = ['parent', 'root', 'partner', 'helper', 'target', 'enemy', 'enemynear', 'helperID', 'targetID', 'enemyID', 'enemynearID', 'playerID']

class RedirectTarget:
    target: Expression
    expr: Expression
    AiLevel: Expression
    Alive: Expression
    Anim: Expression
    AnimElemNo: Callable[..., Expression]
    AnimElemTime: Callable[..., Expression]
    AnimExist: Callable[..., Expression]
    AnimTime: Expression
    AuthorName: Expression
    BackEdgeBodyDist: Expression
    BackEdgeDist: Expression
    CanRecover: Expression
    Command: Expression
    Cond: Callable[..., Expression]
    Const: Callable[..., Expression]
    Ctrl: Expression
    DrawGame: Expression
    Facing: Expression
    FrontEdgeBodyDist: Expression
    FrontEdgeDist: Expression
    GetHitVar: Callable[..., Expression]
    HitCount: Expression
    HitDefAttr: Expression
    HitFall: Expression
    HitOver: Expression
    HitPauseTime: Expression
    HitShakeOver: Expression
    HitVel: PositionExpression
    ID: Expression
    InGuardDist: Expression
    IsHelper: Callable[..., Expression]
    IsHomeTeam: Expression
    Life: Expression
    LifeMax: Expression
    Lose: Expression
    MoveContact: Expression
    MoveGuarded: Expression
    MoveHit: Expression
    MoveReversed: Expression
    MoveType: Expression
    Name: Expression
    NumEnemy: Expression
    NumExplod: Callable[..., Expression]
    NumHelper: Callable[..., Expression]
    NumPartner: Expression
    NumProj: Expression
    NumProjID: Callable[..., Expression]
    NumTarget: Callable[..., Expression]
    P1Name: Expression
    P2BodyDist: PositionExpression
    P2Dist: PositionExpression
    P2Life: Expression
    P2MoveType: Expression
    P2Name: Expression
    P2StateNo: Expression
    P2StateType: Expression
    P3Name: Expression
    P4Name: Expression
    PalNo: Expression
    ParentDist: PositionExpression
    Pos: PositionExpression
    Power: Expression
    PowerMax: Expression
    PrevStateNo: Expression
    ProjCancelTime: Callable[..., Expression]
    ProjContactTime: Callable[..., Expression]
    ProjGuardedTime: Callable[..., Expression]
    ProjHitTime: Callable[..., Expression]
    RootDist: PositionExpression
    ScreenPos: PositionExpression
    SelfAnimExist: Callable[..., Expression]
    StateNo: Expression
    StateType: Expression
    StageVar: Callable[..., Expression]
    TeamMode: Expression
    TeamSide: Expression
    Time: Expression
    Vel: PositionExpression
    Win: Expression
    WinKO: Expression
    WinPerfect: Expression
    WinTime: Expression
    def __init__(self, target: str, expr: Expression | None = None) -> None: ...
    def __getattr__(self, name: str) -> Expression: ...

class RedirectFunction(Protocol):
    def __call__(self, id: Optional[ConvertibleExpression] = ..., /) -> RedirectTarget:
        ...


parent: RedirectTarget
root: RedirectTarget
partner: RedirectTarget
helper: RedirectTarget
target: RedirectTarget
enemy: RedirectTarget
enemynear: RedirectTarget

helperID: RedirectFunction
targetID: RedirectFunction
enemyID: RedirectFunction
enemynearID: RedirectFunction
playerID: RedirectFunction



from mdk.types.specifier import TypeCategory, TypeSpecifier

## these are builtin types which must be available to all characters.
AnyType = TypeSpecifier("any", TypeCategory.BUILTIN) ## note: this is used for DTC and ATC. it's not intended for normal use.
IntType = TypeSpecifier("int", TypeCategory.BUILTIN)
FloatType = TypeSpecifier("float", TypeCategory.BUILTIN)
BoolType = TypeSpecifier("bool", TypeCategory.BUILTIN)
ShortType = TypeSpecifier("short", TypeCategory.BUILTIN)
ByteType = TypeSpecifier("byte", TypeCategory.BUILTIN)
CharType = TypeSpecifier("char", TypeCategory.BUILTIN)
## this is a type for an unquoted string. it basically is still a string,
## but instructs the compiler to not quote its output.
UStringType = TypeSpecifier("string", TypeCategory.BUILTIN)
StringType = TypeSpecifier("string", TypeCategory.BUILTIN)

StateNoType = TypeSpecifier("stateno", TypeCategory.BUILTIN)
AnimType = TypeSpecifier("anim", TypeCategory.BUILTIN)
SoundType = TypeSpecifier("sound", TypeCategory.BUILTIN)
SpriteType = TypeSpecifier("sprite", TypeCategory.BUILTIN)

__all__ = [
    "AnyType", "IntType", "FloatType", "BoolType", "ShortType", "ByteType", "CharType", "StringType",
    "StateNoType", "AnimType", "SoundType", "SpriteType", "UStringType"
]


import ast
import io
from mdk.types.context import StateController as StateController, StateScope as StateScope, TranslationMode as TranslationMode
from mdk.utils.shared import format_bool as format_bool
from mdk.utils.triggers import TriggerAnd as TriggerAnd, TriggerAssign as TriggerAssign, TriggerNot as TriggerNot, TriggerOr as TriggerOr, TriggerPop as TriggerPop, TriggerPush as TriggerPush
from typing import Callable

def write_controller(ctrl: StateController, f: io.TextIOWrapper, locations: bool): ...
def rewrite_function(fn: Callable[..., None], overload_print: bool = True, mode: TranslationMode = TranslationMode.STANDARD) -> Callable[..., None]: ...

class ReplaceLogicalOperators(ast.NodeTransformer):
    location: str
    line: int
    def __init__(self) -> None: ...
    def visit_BoolOp(self, node: ast.BoolOp): ...
    def visit_UnaryOp(self, node: ast.UnaryOp): ...
    def visit_NamedExpr(self, node: ast.NamedExpr): ...
    def visit_If(self, node: ast.If): ...



from typing import NoReturn, Literal
from typing import Callable
from enum import Enum, Flag

from mdk.types.specifier import TypeSpecifier
from mdk.types.context import StateController

__all__ = ['Expression', 'IntExpression', 'ConvertibleExpression', 'TupleExpression']

class Expression:
    exprn: str
    type: TypeSpecifier
    def __init__(self, exprn: str, type: TypeSpecifier) -> None: ...
    def make_expression(self, exprn: str): ...
    def __eq__(self, other: ConvertibleExpression) -> Expression: ... # type: ignore
    def __ne__(self, other: ConvertibleExpression) -> Expression: ... # type: ignore
    def __lt__(self, other: ConvertibleExpression) -> Expression: ...
    def __le__(self, other: ConvertibleExpression) -> Expression: ...
    def __gt__(self, other: ConvertibleExpression) -> Expression: ...
    def __ge__(self, other: ConvertibleExpression) -> Expression: ...
    def __add__(self, other: ConvertibleExpression) -> Expression: ...
    def __sub__(self, other: ConvertibleExpression) -> Expression: ...
    def __mul__(self, other: ConvertibleExpression) -> Expression: ...
    def __truediv__(self, other: ConvertibleExpression) -> Expression: ...
    def __floordiv__(self, other: ConvertibleExpression) -> Expression: ...
    def __mod__(self, other: ConvertibleExpression) -> Expression: ...
    def __pow__(self, other: ConvertibleExpression) -> Expression: ...
    def __rshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __lshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __and__(self, other: ConvertibleExpression) -> Expression: ...
    def __or__(self, other: ConvertibleExpression) -> Expression: ...
    def __xor__(self, other: ConvertibleExpression) -> Expression: ...
    def __radd__(self, other: ConvertibleExpression) -> Expression: ...
    def __rsub__(self, other: ConvertibleExpression) -> Expression: ...
    def __rmul__(self, other: ConvertibleExpression) -> Expression: ...
    def __rtruediv__(self, other: ConvertibleExpression) -> Expression: ...
    def __rfloordiv__(self, other: ConvertibleExpression) -> Expression: ...
    def __rmod__(self, other: ConvertibleExpression) -> Expression: ...
    def __rpow__(self, other: ConvertibleExpression) -> Expression: ...
    def __rlshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __rrshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __rand__(self, other: ConvertibleExpression) -> Expression: ...
    def __ror__(self, other: ConvertibleExpression) -> Expression: ...
    def __rxor__(self, other: ConvertibleExpression) -> Expression: ...
    def __neg__(self) -> Expression: ...
    def __pos__(self) -> Expression: ...
    def __abs__(self) -> Expression: ...
    def __invert__(self) -> Expression: ...
    def __round__(self) -> Expression: ...
    def __trunc__(self) -> Expression: ...
    def __floor__(self) -> Expression: ...
    def __ceil__(self) -> Expression: ...
    def __contains__(self, other) -> Expression: ...
    def __bool__(self) -> Literal[True]: ...

IntExpression: Expression
type ConvertibleExpression = Expression | str | int | float | bool | Callable[..., StateController] | Enum | Flag
type TupleExpression = tuple[ConvertibleExpression, ...]
